/*
 * 2D Polygon Landscape
 *
 * Copyright (c) 2011  Tobias Zwick
 */

#ifndef _LINES_H_
#define _LINES_H_

#include "Vertex.h"
#include "Intersection.h"

/* Line

   Represents a line in 2D space. It can either be a line, a ray or a line
	 segment. Calculations on lines that are parallel to one of the axes are
	 automatically optimized to be faster (about 2x).
 */

class Line
{
	public:
		enum Line_dir { Standard, Horizontal, Vertical };
		enum Line_type { Normal, Ray, Segment };

		Line(const Vertex& first, const Vertex& second, Line_type typ = Normal);
		Line(const Line& l);
		Line();

		bool operator==(const Line& l) const;
		bool operator!=(const Line& l) const;
		Line operator+() const;
		Line operator-() const;

		/* Returns true if this line intersects with the given line. Additionally,
		   the intersection data is stored in intersect. This check is faster for
			 lines that are vertical or horizontal. */
		bool intersects(const Line& other, Intersection &intersect) const;
		/* Returns true if this line intersects with the given line. */
		bool intersects(const Line& other) const;

		/* Returns true if the given point is on the line. */
		bool contains(const Vertex& point) const;

		/* Returns true if the given point is on the line. Additionally, the
		   location of the point on the line is stored in paramVal */
		bool contains(const Vertex& point, double &paramVal) const;	

		/* Determines how a point is situated towards the line. If it is left, this
		   method returns -1, if it is right +1 and if it is on the line 0. Whether
		   this line is a line, line segment or ray is ignored. */
		int getDirection(const Vertex& point) const;

		/* Returns the relative angle towards another line. Possible values are
		   -PI to +PI. */
		double getAngle(const Line& other) const;

		/* Get the line type. Can either be line, line segment or ray. */
		Line_type getType() const;

		/* Set the line type. */
		void setType(Line_type t);

		/* Get the right-hand normal vector. */
		Vertex getNormal() const;

		/* Get the length of the line segment.
		   Even if the type of the line is not a line segment, it is treated
			 as one because to calculate the length of a ray or line wouldn't
			 make any sense (it is =infinity)*/
		double getLength() const;

		/* Get the direction vector of the line (B-A) */
		Vertex getVector() const;

		/* Calculate a point on the line. For line segments, param should be a
		   value between 0 and 1. */
		Vertex getPoint(double param) const;

		/* get start point */
		const Vertex& A() const { return _A; }
		/* set start point */
		void A(const Vertex& val) { _A = val; updateDir(); }
		/* get end point */
		const Vertex& B() const { return _B; }
		/* set end point */
		void B(const Vertex& val) { _B = val; updateDir(); }

	private:
		Vertex _A,_B;

		Line_dir line;
		Line_type type;
		
		bool paramBounds(double param) const;
		void updateDir();
		bool sameLineIntersect(const Line& other, Intersection &intersect) const;
		bool simpleIntersect(const Line& other, Intersection &intersect) const;
};

#endif

