#include "PolygonClipper.h"

#include <vector>
#include <stack>
#include <cassert>
#define _USE_MATH_DEFINES
#include <cmath>

#include "Intersection.h"
#include "ringvec.h"

PolygonClipper::PolygonClipper(const Polygon &s, const Polygon &c)
: id(0),subject_polygon(s),clip_polygon(c),has_self_intersections_in_a(false),has_self_intersections_in_b(false)
{
	// check bounding boxes first
	if(!subject_polygon.insideBoundingBox(clip_polygon))
	{
		outside_bounding_box = true;
		return;
	}
	outside_bounding_box = false;

	// fill up augmented polygons
	for(std::vector<Vertex>::const_iterator it = subject_polygon.vertices().begin(); it != subject_polygon.vertices().end(); ++it)
		polygon_a.push_back(PolyVertex(*it,id++));

	for(std::vector<Vertex>::const_iterator it = clip_polygon.vertices().begin(); it != clip_polygon.vertices().end(); ++it)
		polygon_b.push_back(PolyVertex(*it,id++));

	has_intersections = calculateIntersections();

	// look if there are any self-intersections
	// TODO: incomplete implementation. uncomment when the issues
	// there have been solved
	if(!self_intersections_a.empty())
	{
		//polygons_of_a = splitUpAtSelfIntersections(polygon_a,self_intersections_a);
		has_self_intersections_in_a = true;
	}
	if(!self_intersections_b.empty())
	{
		//polygons_of_b = splitUpAtSelfIntersections(polygon_b,self_intersections_b);
		has_self_intersections_in_b = true;
	}
}

PolygonClipper::~PolygonClipper()
{
}

/* +++++++++++++++++++++++++++++++++ Main ++++++++++++++++++++++++++++++++++ */

Clip_result PolygonClipper::clip(Clipping_operation op, std::list<Polygon *> &result)
{
	// check bounding boxes first
	if(outside_bounding_box)
		return standardClipResult(op);

	// if there are no intersections: tag all the same or cancel...
	if(!has_intersections)
	{
		Poly_tag tag_for_a = None;
		Poly_tag tag_for_b = None;
		// first determine their orientation to each other:
		// one point of other polygon is inside this polygon
		if(subject_polygon.contains(clip_polygon.vertices().front()))
		{
			tag_for_a = Outside;
			tag_for_b = Inside;
		}
		// one point of this polygon is inside other polygon
		else if(clip_polygon.contains(subject_polygon.vertices().front()))
		{
			tag_for_a = Inside;
			tag_for_b = Outside;
		}
		// if there are no intersections and any point of the polygon b and a is
		// not inside the other, it means the polygons do not actually intersect.
		// We can return standard results
		else
		{
			return standardClipResult(op);
		}
		// Otherwise - depending on the clipping operation - we can a shortcut
		// for some of the clipping operations here and return unmodified polygons
		Clip_result enclosed = enclosedClipResult(op,tag_for_a);
		if(enclosed != Default)
			return enclosed;

		// otherwise, tag all the edges the same and skip the
		// regular tagging
		for(polyvertex_it i = polygon_a.begin(); i != polygon_a.end(); ++i)
			i->tag = tag_for_a;
		for(polyvertex_it i = polygon_b.begin(); i != polygon_b.end(); ++i)
			i->tag = tag_for_b;
	}
	// there are several subject and clip polygons now... :-(
	else if(has_self_intersections_in_a || has_self_intersections_in_b)
	{
		// TODO: it can also be that after the splitting, the polygons_of_a is
		// empty because it only consisted of invalid polygons!

		// TODO: finish this implementation (until then, we return SubjectPolygon
		// on "error")
		//return handleMultipleClippingOperations(op,result);
		return SubjectPolygon;
	}
	// tag the polygons
	else
	{
		Poly_tag tag = tagPolygons();

		// shortcuts:
		// Polygon A does only consist of shared edges (A clip A)
		if(tag == Shared)
		{
			// nothing
			if(op == Difference || op == Exclusion)
				return Nothing;
			// subject polygon
			if(op == Union || op == Intersect)
				return SubjectPolygon;
		}
		// Polygon A does only consist of outside edges (polygons
		// max. only touch each other at intersection points)
		else if(tag == Outside)
		{
			return standardClipResult(op);
		}
		// Polygon A does consist only of outside and reverse-shared
		// edges (polygons only touch each other)
		else if(tag == ReverseShared)
		{
			if(op == Difference)
				return SubjectPolygon;
			if(op == Intersect)
				return Nothing;
		}
		// Polygon A does consist only of inside and shared edges
		// (polygons only touch each other but one is contained in
		// the other)
		else if(tag == Inside)
		{
			Clip_result enclosed = enclosedClipResult(op,tag);
			if(enclosed != Default)
				return enclosed;
		}
	}

	// then, put the polygons together...
	std::list<Polygon *> clipResult;
	constructPolygons(op,clipResult);

	// for every hole: split polygons...
	eliminateHoles(clipResult);

	// at last, put result in the result list
	result.splice(result.end(),clipResult);

	return Default;
}

Clip_result PolygonClipper::enclosedClipResult(Clipping_operation op, Poly_tag tag_for_a) const
{
		if(op == Difference)
		{
			// Difference: if A is smaller, return nothing
			if(tag_for_a == Inside)
				return Nothing;
		}
		else if(op == Union)
		{
			// Union: return only the larger one
			if(tag_for_a == Inside) return ClipPolygon;
			else return SubjectPolygon;
		}
		else if(op == Intersect)
		{
			// Intersection: return only the smaller one
			if(tag_for_a == Outside) return ClipPolygon;
			else return SubjectPolygon;
		}
		return Default;
}

Clip_result PolygonClipper::standardClipResult(Clipping_operation op) const
{
	// these are the standard clip results for the different operations if the
	// two polygons do not intersect:
	if(op == Difference)
		return SubjectPolygon;
	if(op == Union)
		return BothPolygons;
	if(op == Exclusion)
		return BothPolygons;
	if(op == Intersect)
		return Nothing;
	return Default;
}


/* +++++++++++++++++++++++++ Step 1: Intersections +++++++++++++++++++++++++ */

bool PolygonClipper::calculateIntersections()
{
	/*
	Step 1: calculate intersections and insert them in the tagged
	polygons. Since we need to check intersections of every line of one
	polygon with every line of the other polygon, this is expensive but
	there is no other way to do that
	*/	
	bool hasIntersections = false;

	// for every line of Polygon A
	for(polyvertex_it i = polygon_a.begin(); i != polygon_a.end(); ++i)
	{
		// for every line of Polygon B
		for(polyvertex_it j = polygon_b.begin(); j != polygon_b.end(); ++j)
		{
			if(handleIntersection(i,j))
				hasIntersections = true;
		}
	}
	return hasIntersections;
}

bool PolygonClipper::handleIntersection(polyvertex_it &i, polyvertex_it &j)
{

	Vertex P = decreaseIt(j,polygon_b)->data;
	Vertex Q = j->data;

	Vertex A = decreaseIt(i,polygon_a)->data;
	Vertex B = i->data;

	Line edge_in_a(A,B,Line::Segment);
	Line edge_in_b(P,Q,Line::Segment);

	Intersection intersect;

	// not found intersection
	if(!edge_in_a.intersects(edge_in_b,intersect)) return false;

	// for points: insert if not already there
	if(!intersect.isLineIntersection())
	{
		Vertex p = *(intersect.point);
		polyvertex_it posA, posB;

		handlePointIntersection(p,edge_in_a,i,polygon_a,posA);
		handlePointIntersection(p,edge_in_b,j,polygon_b,posB);

		// if the point of the intersection has been created at it's position without
		// rounding, memorize the intersection
		if(intersect.exact)
		{
			assert(posA->data == posB->data);
			Vertex key(posA->data);
			// link from PolyVertexLink to iterators at intersection
			intersections[key].positionsInA.insert(posA);
			intersections[key].positionsInB.insert(posB);
			// link from PolyVertices to which the iterators link
			// back to the PolyVertexLink
			posA->link = &intersections[key];
			posB->link = &intersections[key];
		}
		// otherwise, we have to first check for more intersections because the shape
		// of the polygon just changed.
		else
		{
			checkForNewIntersections(posA, posB);
		}
	}
	// for lines 
	else
	{
		Line l = *(intersect.line);

		polyvertex_it posA1;
		polyvertex_it posA2;
		polyvertex_it posB1;
		polyvertex_it posB2;

		handleLineIntersection(l, edge_in_a, i, polygon_a, posA1, posA2);
		handleLineIntersection(l, edge_in_b, j, polygon_b, posB1, posB2);

		// (if we have a line intersection, the two points are always created without rounding
		// imprecision. Thats why we memorize the intersection directly and don't check recursively)

		// insert links from PolyVertexLink to (iterators to) PolyVertices and back.
		intersections[posA1->data].positionsInA.insert(posA1);
		intersections[posA2->data].positionsInA.insert(posA2);
		intersections[posB1->data].positionsInB.insert(posB1);
		intersections[posB2->data].positionsInB.insert(posB2);
		posA1->link = &intersections[posA1->data];
		posA2->link = &intersections[posA2->data];
		posB1->link = &intersections[posB1->data];
		posB2->link = &intersections[posB2->data];
	}

	return true;
}

void PolygonClipper::handlePointIntersection(	const Vertex &intersectPoint,	const Line &line,
																							polyvertex_it &insert, std::list<PolyVertex> &poly,
																							polyvertex_it &pos) 
{
	// first point of the line segment counts as the last point of the
	// last line segment
	if(intersectPoint == line.A())
		pos = decreaseIt(insert,poly);
	else if(intersectPoint == line.B())
		pos = insert;
	// if intersection on line, insert it
	else
		insert = pos = poly.insert(insert,PolyVertex(intersectPoint,id++));
}

void PolygonClipper::handleLineIntersection(	const Line &intersectLine, const Line &line,
																							polyvertex_it &insert, std::list<PolyVertex> &poly,
																							polyvertex_it &pos1, polyvertex_it &pos2) 
{
	// make intersection line go into the same direction as this line
	//(makes if-queries easier)
	bool oppositeDirection = line.getVector()*intersectLine.getVector() < 0;

	Vertex A = line.A();
	Vertex B = line.B();

	Vertex p = oppositeDirection ? intersectLine.B() : intersectLine.A();
	Vertex q = oppositeDirection ? intersectLine.A() : intersectLine.B();

	polyvertex_it i = insert;

	// insert both points
	if(p != A && q != B)
	{
		i = pos1 = poly.insert(insert,PolyVertex(p,id++));
		pos2 = poly.insert(insert,PolyVertex(q,id++));
	}
	// both points already exist
	else if(p == A && q == B)
	{
		pos1 = decreaseIt(insert,poly);
		pos2 = insert;
	}
	// one of the points must be inserted
	else if(p == A)
	{
		pos1 = decreaseIt(insert,poly);
		i = pos2 = poly.insert(insert,PolyVertex(q,id++));
	}
	else if(q == B)
	{
		i = pos1 = poly.insert(insert,PolyVertex(p,id++));
		pos2 = insert;
	}

	// The current polyvertex_it must be set to the first inserted vertex
	// so that no edges are forgotten during the intersection checks.
	// If none have been inserted, the current polyvertex_it it not changed.
	insert = i;
}



bool PolygonClipper::handleSelfIntersection(	polyvertex_it &it, polyvertex_it &jt, bool forB	)
{
	if(it == jt) return false;

	std::list<PolyVertex> &poly = forB ? polygon_b : polygon_a;

	polyvertex_it it_before = decreaseIt(it,poly);
	polyvertex_it jt_before = decreaseIt(jt,poly);

	Line line1(it_before->data,it->data,Line::Segment);
	Line line2(jt_before->data,jt->data,Line::Segment);

	Intersection intersect;
	if(!line1.intersects(line2,intersect)) return false;

	// every line intersects with it's successor and the one behind.
	// we can ignore that
	if(!intersect.isLineIntersection())
	{
		if(increaseIt(it,poly) == jt && intersect.thisParam == 1.0)
			return false;
		if(it_before == jt && intersect.thisParam == 0.0)
			return false;
	}

	std::map<Vertex,PolyVertexLink,VertexComp> &self_intersections = forB ? self_intersections_b : self_intersections_a;

	// line intersection
	if(intersect.isLineIntersection())
	{
		polyvertex_it line1_pos1, line1_pos2, line2_pos1, line2_pos2;

		handleLineIntersection(*intersect.line,line1,it,poly,line1_pos1, line1_pos2);
		handleLineIntersection(*intersect.line,line2,jt,poly,line2_pos1, line2_pos2);

		// insert links from PolyVertexLink to (iterators to) PolyVertices and back.
		// only insert if these are really two different PolyVertices
		if(line1_pos1 != line2_pos2)
		{
			self_intersections[line1_pos1->data].positionsInA.insert(line1_pos1);
			self_intersections[line2_pos2->data].positionsInA.insert(line2_pos2);
			line1_pos1->selflink = &self_intersections[line1_pos1->data];
			line2_pos2->selflink = &self_intersections[line2_pos2->data];
		}
		// only insert if these are really two different PolyVertices
		if(line1_pos2 != line2_pos1)
		{
			self_intersections[line1_pos2->data].positionsInA.insert(line1_pos2);
			self_intersections[line2_pos1->data].positionsInA.insert(line2_pos1);
			line1_pos2->selflink = &self_intersections[line1_pos2->data];
			line2_pos1->selflink = &self_intersections[line2_pos1->data];
		}
	}
	// point intersection
	else
	{
		polyvertex_it line1_pos, line2_pos;
		
		handlePointIntersection(*intersect.point, line1, it, poly, line1_pos);
		handlePointIntersection(*intersect.point, line2, jt, poly, line2_pos);

		assert(line1_pos->data == line2_pos->data);
		Vertex key(line1_pos->data);
		// link from PolyVertexLink to iterators at intersection
		self_intersections[key].positionsInA.insert(line1_pos);
		self_intersections[key].positionsInA.insert(line2_pos);
		// link from PolyVertices to which the iterators link
		// back to the PolyVertexLink
		line1_pos->selflink = &self_intersections[key];
		line2_pos->selflink = &self_intersections[key];

		if(!intersect.exact)
		{
			if(!forB)
			{
				polyvertex_it nada =  polygon_b.end();
				checkForNewIntersections(	line1_pos, nada);
				checkForNewIntersections(	line2_pos, nada);
			}
			else
			{
				polyvertex_it nada =  polygon_a.end();
				checkForNewIntersections(	nada, line1_pos );
				checkForNewIntersections(	nada, line2_pos );
			}
		}
	}
	return true;
}



void PolygonClipper::checkForNewIntersections(	polyvertex_it &posA, polyvertex_it &posB )
{
	if(posA != polygon_a.end())
	{
		polyvertex_it n_posA = increaseIt(posA,polygon_a);

		// check for self-intersections in polygon A
		for(polyvertex_it c = polygon_a.begin(); c != polygon_a.end(); ++c)
		{
			handleSelfIntersection(posA, c, false);
			handleSelfIntersection(n_posA, c, false);
		}
		// check for intersections between the 2 new edges in polygon A with any
		// edges in polygon B
		for(polyvertex_it c = polygon_b.begin(); c != polygon_b.end(); ++c)
		{
			handleIntersection(posA,c);
			handleIntersection(n_posA,c);
		}
	}
	if(posB != polygon_b.end())
	{
		polyvertex_it n_posB = increaseIt(posB,polygon_b);

		// check for self-intersections in polygon B
		for(polyvertex_it c = polygon_b.begin(); c != polygon_b.end(); ++c)
		{
			handleSelfIntersection(posB, c, true);
			handleSelfIntersection(n_posB, c, true);
		}
		// same with other
		for(polyvertex_it c = polygon_a.begin(); c != polygon_a.end(); ++c)
		{
			handleIntersection(c,posB);
			handleIntersection(c,n_posB);
		}
	}
}


std::list<std::list<PolygonClipper::PolyVertex> > PolygonClipper::splitUpAtSelfIntersections(std::list<PolyVertex> &poly, 
																			std::map<Vertex,PolyVertexLink,VertexComp> &self_intersections) const
{
	std::list<std::list<PolyVertex> > polygons;
	std::stack<polyvertex_it> polystack;

	// start at one self intersection
	PolyVertexLink &pvl(self_intersections.begin()->second);
	polyvertex_it start = *pvl.positionsInA.begin();
	polyvertex_it it = start;
	polystack.push(start);
	bool passedStart = false;

	do
	{
		// goto marker to start of loop
		start_of_loop:

		// walk until next self-intersection has been found
		polyvertex_it next_it = increaseIt(it,poly);
		if(next_it == poly.begin()) passedStart = true;

		if(next_it->selflink)
		{
			bool same_point = next_it->data == polystack.top()->data;
			// if we reached a self-intersection which is at the same point as the one
			// we started with, finalize polygon 
			if(same_point)
			{
				polygons.push_back(std::list<PolyVertex>());
				std::list<PolyVertex> &mypoly = polygons.back();
				// ...from top of stack ... to ... current vertex
				polyvertex_it polygon_start = polystack.top();

				// finish up polygon (the next vertex is the vertex we started with,
				// this can only mean that we have gone round completely)
				if(polygon_start == next_it)
				{
					mypoly.splice(mypoly.begin(),poly);
					break;
				}
				// otherwise splice normally. But take care to go over the end
				else if(!passedStart)
				{
					mypoly.splice(mypoly.begin(),poly,polygon_start,next_it);
				}
				else
				{
					mypoly.splice(mypoly.begin(),poly,polygon_start,poly.end());
					mypoly.splice(mypoly.end(),poly,poly.begin(),next_it);
					// TODO: sometimes passedStart needs to be set to false here...
					// but when? Or is passedStart just set to true too early?
				}
				polystack.pop();
				// our start vertex has been popped
				if(polystack.empty())
				{
					passedStart = false;
					polystack.push(next_it);
					it = start = next_it;
					goto start_of_loop;
				}
			}
			// otherwise, there must be a "pimple" on that one. Push a new start vertex
			// to the stack
			else
			{
				polystack.push(next_it);
			}
		}
		it = next_it;

	} while(it != start);

	assert(polygons.size() > 1);
	assert(poly.empty());

	// filter out polygons with zero size (can only be: two vertices cause they are all separated)
	for(std::list<std::list<PolyVertex> >::iterator pit = polygons.begin(); pit != polygons.end(); )
	{
		assert(pit->size() > 1);
		if(pit->size() == 2)
			pit = polygons.erase(pit);
		else
			++pit;
	}

	// return list of list of polyvertices
	return polygons;
}


/* ++++++++++++++++++++++++++++ Step 2: Tagging ++++++++++++++++++++++++++++ */

PolygonClipper::Poly_tag PolygonClipper::tagPolygons()
{
	/*
		Step 2: Tag polygons. For each line of both polygons we need to find out
		if they are inside, outside, shared or reverse shared line segments...
	*/

	bool aOnlyOutside = true, aOnlyShared = true,
	     aOnlyOutsideOrReverseShared = true, aOnlyInsideOrShared = true;

	// for all intersections between the two polygons...
	for(polyvertexlink_it it = intersections.begin(); it != intersections.end(); ++it)
	{

		PolyVertexLink &pvl(it->second);
		for(intersect_pos_it at = pvl.positionsInA.begin(); at != pvl.positionsInA.end(); ++at)
		{
			Poly_tag tag = tagForIntersection(*at,false);

			// if all edges of the subject polygon are tagged the same, certain
			// shortcuts can be made (see below)
			if(tag == None) continue;
			if(tag != Shared)
			{
				aOnlyShared = false;
				if(tag != Inside)
					aOnlyInsideOrShared = false;
			}
			if(tag != Outside)
			{
				aOnlyOutside = false;
				if(tag != ReverseShared)
					aOnlyOutsideOrReverseShared = false;
			}
		}
		for(intersect_pos_it bt = pvl.positionsInB.begin(); bt != pvl.positionsInB.end(); ++bt)
		{
			tagForIntersection(*bt,true);
		}
	}

#ifndef NDEBUG

	// all edges must be tagged at this point
	for(polyvertex_it it = polygon_a.begin(); it != polygon_a.end(); ++it)
		assert(it->tag != PolygonClipper::None);
	for(polyvertex_it jt = polygon_b.begin(); jt != polygon_b.end(); ++jt)
		assert(jt->tag != PolygonClipper::None);

#endif

	// shortcuts for certain tagggings of polygon A
	if(aOnlyShared) return Shared;
	if(aOnlyOutside) return Outside;
	if(aOnlyOutsideOrReverseShared) return ReverseShared;
	if(aOnlyInsideOrShared) return Inside;
	return None;
}

PolygonClipper::Poly_tag PolygonClipper::tagForIntersection(const polyvertex_it &p, bool forB)
{
	polyvertex_it nextP = forB ? increaseIt(p,polygon_b) : increaseIt(p,polygon_a);

	// skip if for this intersection, we already tagged the edges
	// (happens when at this point, we have multiple intersections with the other polygon)
	if (p->tag != None)
		return None;

	// find the two edges of the other polygon which are closest to the line which interests us now
	// normally there are just two edges but theoretically, there can be several for weakly simple
	// polygons
	Line thisLine(p->data,nextP->data);

	Line prevLine;
	Line nextLine;

	std::set<polyvertex_it,PolyVertexItComp> &verticesInOtherPolygon = forB ? p->link->positionsInA : p->link->positionsInB;
	std::list<PolyVertex> &poly = forB ? polygon_a : polygon_b;

	calculateClosestEdgesToLine(thisLine, prevLine, nextLine, verticesInOtherPolygon, poly, forB);

	// determine if the current line is inside or outside the other polygon by
	// looking at the two surrounding edges of the other polygons
	Poly_tag tag = getLineOrientation(prevLine,nextLine,thisLine);

	// now tag the polygon till the next intersection is found
	polyvertex_it pit = p;
	do {
		pit->tag = tag;
		// we increase the iterator with this function instead of ++ because
		// a polygon is a ring - if the end of the list of vertices has been
		// reached, he should just start at the front again.
		pit = forB ? increaseIt(pit,polygon_b) : increaseIt(pit,polygon_a);
	} while(!pit->link);

	return tag;
}

void PolygonClipper::calculateClosestEdgesToLine(const Line& line, Line &prevLine, Line &nextLine,
																					std::set<polyvertex_it,PolyVertexItComp> &verticesInOtherPolygon,
																					std::list<PolyVertex> &otherPolygon, bool forB) const
{
		double min_prev_angle = 10.0, max_prev_angle = -10.0;
		double min_next_angle = 10.0, max_next_angle = -10.0;

		Line min_prev, max_prev, min_next, max_next;

		// for all vertices in the other polygon which are referenced by the intersection point from A...
		for(intersect_pos_it jt = verticesInOtherPolygon.begin(); jt != verticesInOtherPolygon.end(); ++jt )
		{
			polyvertex_it other = (*jt);

			// get the previous and next vertex of the other polygon...
			polyvertex_it other_previous = decreaseIt(other,otherPolygon);
			polyvertex_it other_next = increaseIt(other,otherPolygon);

			Vertex A(other_previous->data);
			Vertex P(other->data);
			Vertex B(other_next->data);
			Line curr_prevLine(A,P);
			Line curr_nextLine(P,B);

			// and get their angles towards the line
			double curr_prev_angle = line.getAngle(curr_prevLine);
			double curr_next_angle = line.getAngle(curr_nextLine);
			if(curr_next_angle < 0) curr_next_angle += 2.0*M_PI;

			// determine previous line with minimum angle
			if(curr_prev_angle < min_prev_angle)
			{
				min_prev_angle = curr_prev_angle;
				min_prev = curr_prevLine;
			}
			// determine previous line with maximum angle
			if(curr_prev_angle > max_prev_angle)
			{
				max_prev_angle = curr_prev_angle;
				max_prev = curr_prevLine;
			}
			// determine next line with minimum angle
			if(curr_next_angle < min_next_angle)
			{
				min_next_angle = curr_next_angle;
				min_next = curr_nextLine;
			}
			// determine next line with maximum angle
			if(curr_next_angle > max_next_angle)
			{
				max_next_angle = curr_next_angle;
				max_next = curr_nextLine;
			}
		}
		// of the two sets of lines selected, select the set which encloses the
		// the smallest angle
		double max_prev_min_next_angle = getEnclosingAngle(max_prev_angle, min_next_angle);
		double min_prev_max_next_angle = getEnclosingAngle(min_prev_angle, max_next_angle);

		if(max_prev_min_next_angle < min_prev_max_next_angle)
		{
			prevLine = max_prev;
			nextLine = min_next;
		}
		else
		{
			prevLine = min_prev;
			nextLine = max_next;
		}
}

double PolygonClipper::getEnclosingAngle(double prev, double next) const
{
	if(prev + M_PI > next)
		return M_PI - prev + next;
	else
		return M_PI + prev - next + M_PI*2.0;
}

PolygonClipper::Poly_tag PolygonClipper::getLineOrientation(const Line &prevLine, const Line &nextLine,
																														const Line &line) const
{
		// its on the line itself?
		// same direction as the other lines
		if(line.B() == nextLine.B()) return Shared;
		// opposite direction as the other lines
		if(line.B() == prevLine.A()) return ReverseShared;

		bool line1inside = prevLine.getDirection(line.B()) <= 0;
		bool line2inside = nextLine.getDirection(line.B()) <= 0;
		bool linesconcave =  prevLine.getAngle(nextLine) > 0;

		Poly_tag result = Outside;

		// if the other lines are concave, the line must be on the left side of
		// either of the two
		if(linesconcave)
		{
			if(line1inside || line2inside)
				result = Inside;
		}
		// otherwise must be on the left side of both
		else
		{
			if(line1inside && line2inside)
				result = Inside;
		}
		return result;
}


/* +++++++++++++++++++++ Step 3: Construct polygon(s) ++++++++++++++++++++++ */

void PolygonClipper::constructPolygons(Clipping_operation op, std::list<Polygon*> &clipResult)
{
	// the polygon(s) are constructed in this list, one vertex after one
	// another is added during the process.
	std::list<Vertex> construct;

	// all vertex_links are iterated
	for(polyvertexlink_it it = intersections.begin(); it != intersections.end(); ++it)
	{
		PolyVertexLink *link = &(it->second);

		// if the vertex_link has already been visited in below loop: to next intersection
		if(link->visited) continue;

		assert(construct.empty());

		// in one cycle of this loop, we select one path and walk along it until
		// we reach the next intersection. We do that until we reach an
		// intersection which has already been visited
		do {

			// select the egible paths we can walk. Whats egible and what's not is
			// dependent on the clipping operation we chose.
			std::list<PolyPathInfo> availPaths = getAllAvailablePaths(link,op);
			// no available paths? skip
			if(availPaths.empty())
			{
				link->visited = true;
				continue;
			}
			// choose the path which goes - relative to the last line added to
			// the polygon in construction - into the leftmost direction
			// If this will be the first line of the polygon, choose any path
			std::list<PolyPathInfo>::iterator best = availPaths.begin();


			if(construct.size() > 0 && availPaths.size() > 1)
			{
				double bestAngle = M_PI;
				// last line: A = last vertex in construct; B = any walker data of the available paths
				Line last(*construct.rbegin(), best->walker->data);

				// this loop should iterate over 1-4 paths for simple polygons (and any number for weakly simple polygons)
				for(std::list<PolyPathInfo>::iterator tri = availPaths.begin(); tri != availPaths.end(); ++tri)
				{
					// walk along this path to get the line
					Vertex A(tri->walker->data), B;
					if(!tri->reverse)
						B = increaseIt(tri->walker,tri->path)->data;
					else
						B = decreaseIt(tri->walker,tri->path)->data;
					Line line(A,B);
					double angle = last.getAngle(line);

					if(angle <= bestAngle)
					{
						bestAngle = angle;
						best = tri;
					}
				}
			}
			// left-most path selected
			PolyPathInfo *c = &(*best);

			// mark selected path as visited: tag the edge as None so that it is not
			// found again later and if it was the only possible path, tag the
			// vertex_link as visited
			if(c->reverse)
				decreaseIt(c->walker,c->path)->tag = None;
			else
				c->walker->tag = None;
				
			if(availPaths.size() == 1)
				link->visited = true;

			// walk along the path until next intersection and put all vertices that
			// are on the way into construct
			do
			{
				construct.push_back(c->walker->data);
				if(c->reverse)
					c->walker = decreaseIt(c->walker,c->path);
				else
					c->walker = increaseIt(c->walker,c->path);
			}
			while(!c->walker->link);

			// new intersection reached
			link = c->walker->link;
			
			// if we reached the starting point of the polygon we are constructing,
			// finalize polygon and clear construct.
			if(c->walker->data == construct.front())
			{
				Polygon *poly = new Polygon(construct);
				clipResult.push_back(poly);
				construct.clear();
			}

		} while(!link->visited);

	}
	// no intersections at all. At this point, this can only occur
	// for difference or exclusion. The other operations return earlier
	if(intersections.empty())
	{
		assert(op == Difference || op == Exclusion);
		// construct polygons: Inside = reverse
		//                     Outside = normal
		Polygon *first;
		Polygon *second;
		// polygon_a == subject polygon
		if(polygon_a.front().tag == Outside)
			first = new Polygon(subject_polygon);
		else
			first = new Polygon(-subject_polygon);

		// polygon_b == clip polygon
		if(polygon_b.front().tag == Outside)
			second = new Polygon(clip_polygon);
		else
			second = new Polygon(-clip_polygon);

		clipResult.push_back(first);
		clipResult.push_back(second);
	}
}


std::list<PolygonClipper::PolyPathInfo> PolygonClipper::getAllAvailablePaths(	PolyVertexLink* link,
																																							Clipping_operation op)
{
	std::list<PolyPathInfo> availPaths;

	// first look at all paths in A (normally, it is just one but for weakly simple
	// polygons, theoretically there can be any number)
	for(intersect_pos_it at = link->positionsInA.begin(); at != link->positionsInA.end(); ++at)
	{
		polyvertex_it walker = *at;
		Poly_tag next = walker->tag;
		Poly_tag prev = decreaseIt(walker,polygon_a)->tag;

		if (op == Difference)
		{
			if(next == Outside || next == ReverseShared)
				availPaths.push_back(PolyPathInfo(polygon_a,walker,false));
		}
		else if(op == Union)
		{
			if(next == Outside || next == Shared)
				availPaths.push_back(PolyPathInfo(polygon_a,walker,false));
		}
		else if(op == Intersect)
		{
			if(next == Inside || next == Shared)
				availPaths.push_back(PolyPathInfo(polygon_a,walker,false));
		}
		else if(op == Exclusion)
		{
			if(next == Outside)
				availPaths.push_back(PolyPathInfo(polygon_a,walker,false));
			if(prev == Inside)
				availPaths.push_back(PolyPathInfo(polygon_a,walker,true));
		}
	}
	// then, for all in B
	for(intersect_pos_it bt = link->positionsInB.begin(); bt != link->positionsInB.end(); ++bt)
	{
		polyvertex_it walker = *bt;
		Poly_tag next = walker->tag;
		Poly_tag prev = decreaseIt(walker,polygon_b)->tag;

		if (op == Difference)
		{
			if(prev == Inside)
				availPaths.push_back(PolyPathInfo(polygon_b,walker,true));
		}
		else if(op == Union)
		{
			if(next == Outside)
				availPaths.push_back(PolyPathInfo(polygon_b,walker,false));
		}
		else if(op == Intersect)
		{
			if(next == Inside)
				availPaths.push_back(PolyPathInfo(polygon_b,walker,false));
		}
		else if(op == Exclusion)
		{
			if(next == Outside)
				availPaths.push_back(PolyPathInfo(polygon_b,walker,false));
			if(prev == Inside)
				availPaths.push_back(PolyPathInfo(polygon_b,walker,true));
		}
	}
	return availPaths;
}


/* +++++++++++++++++++++++ Step 4: Eliminate holes +++++++++++++++++++++++++ */

void PolygonClipper::eliminateHoles(std::list<Polygon *> &clipResult) const
{
	// 1. determine which polygons are the holes and put them in a separate list (ccw defined)
	std::list<Polygon *> holes;
	for(std::list<Polygon *>::iterator it = clipResult.begin(); it != clipResult.end(); it++)
	{
		Polygon *resultPolygon = *it;
		if(resultPolygon->size() < 0)
		{
			// NOTE: a new polygon wouldn't have to be created here if our clipping algorithm
			// worked with holes.
			Polygon *ccwHole = new Polygon(-(*resultPolygon));
			holes.push_back(ccwHole);
			it = clipResult.erase(it);
			delete resultPolygon;
			if(it != clipResult.begin()) --it;
			if(clipResult.empty()) break;
		}
	}
	// 2. for every hole there is, split all polygons
	for(std::list<Polygon *>::iterator holesit = holes.begin(); holesit != holes.end(); ++holesit)
	{
		Polygon *hole = *holesit;

		// determine split line (currently: just first edge of hole)
		std::vector<Vertex>::const_iterator v = hole->vertices().begin();
		Vertex first = *v;
		Vertex second = *(++v);
		Line splitLine(first,second,Line::Normal);

		std::list<Polygon *> splitResult;
		for(std::list<Polygon *>::iterator polyit = clipResult.begin(); polyit != clipResult.end();)
		{
			Polygon *polygon = *polyit;
			// polygon is split -> result is put into splitResult 
			if(polygon->split(splitLine,splitResult) != SubjectPolygon)
			{
				// delete old polygon
				polyit = clipResult.erase(polyit);
				delete polygon;
				if(clipResult.empty()) break;
			}
			else ++polyit;
		}
		// put the results into clipResult
		clipResult.splice(clipResult.end(),splitResult);
	}

	// 3. subtract every hole from all of the split up polygons
	for(std::list<Polygon *>::iterator holesit = holes.begin(); holesit != holes.end(); ++holesit)
	{
		Polygon *hole = *holesit;
		std::list<Polygon *> subtractHoleResult;
		for(std::list<Polygon *>::iterator polyit = clipResult.begin(); polyit != clipResult.end(); )
		{
			Polygon *polygon = *polyit;
			Clip_result clipHoleResult = polygon->clip(*hole,Difference,subtractHoleResult);
			// the polygon does not remain untouched...
			if(clipHoleResult != SubjectPolygon)
			{
				// delete old polygon
				polyit = clipResult.erase(polyit);
				delete polygon;
				if(clipResult.empty()) break;
			}
			else polyit++;
		}
		clipResult.splice(clipResult.end(),subtractHoleResult);
	}

	#ifndef NDEBUG

		// there must be no holes in the output
		for(std::list<Polygon *>::const_iterator it = clipResult.begin(); it != clipResult.end(); ++it)
			assert((*it)->size() > 0);

	#endif
}

/* ++++++++++++++++++ Special: handle multiple intersectins ++++++++++++++++ */

Clip_result PolygonClipper::handleMultipleClippingOperations(Clipping_operation op, std::list<Polygon *> &result)
{
	// first put the remaining polygon_a/polygon_b into the list
	if(!polygon_a.empty())
	{
		assert(polygon_b.empty());
		// this way, the list is not copied but just moved
		polygons_of_a.push_back(std::list<PolyVertex>());
		polygons_of_a.back().splice(polygons_of_a.back().end(),polygon_a);
	}
	if(!polygon_b.empty())
	{
		assert(polygon_a.empty());
		// this way, the list is not copied but just moved
		polygons_of_b.push_back(std::list<PolyVertex>());
		polygons_of_b.back().splice(polygons_of_b.back().end(),polygon_b);
	}

	// NOTE: for now, we will just create new polygons out of this and start the 
	// polygon clipping operation(s) anew. However, if we do it properly, we 
	// could leave out the first step (finding, inserting and memorizing 
	// intersections) because we already did this at this point. However this
	// would require to make the tagging/constructing algorithm either work with
	// a "global" list of intersections or to split up the list of intersections
	// for every split up polygon. The problem is (of a global list), that it is
	// not possible to get the information to which list an iterator in the list
	// of intersections points to
	// however, it is also not so trivial to pick the intersection map to pieces

	// create the polygons
	std::list<Polygon *> polys_a;
	for(std::list<std::list<PolyVertex> >::iterator it = polygons_of_a.begin(); it != polygons_of_a.end(); ++it)
		polys_a.push_back(createPolygon(*it));

	std::list<Polygon *> polys_b;
	for(std::list<std::list<PolyVertex> >::iterator it = polygons_of_b.begin(); it != polygons_of_b.end(); ++it)
		polys_a.push_back(createPolygon(*it));

	// first, handle own holes
	eliminateHoles(polys_a);
	eliminateHoles(polys_b);

	// now, do the operation:

	// for difference, subtract everything in B from each in A
	if(op == Difference)
	{
		for(std::list<Polygon *>::iterator ait = polys_a.begin(); ait != polys_a.end(); )
		{
			bool clipped = false;

			for(std::list<Polygon *>::iterator bit = polys_b.begin(); bit != polys_b.end(); ++bit)
			{
				std::list<Polygon *> clipResult;
				Clip_result cr = (*ait)->clip(**bit,Difference,clipResult);
				// if something has been clipped, remove the original subject polygon
				// and replace it with the clip result in the list
				if(cr == Default || cr == Nothing)
				{
					delete *ait;
					ait = polys_a.erase(ait);
					clipped = true;
					polys_a.splice(polys_a.end(),clipResult);
					break;
				}
			}
			if(!clipped) ++ait;
		}
		// clear polygons in polys_b
		for(std::list<Polygon *>::iterator bit = polys_b.begin(); bit != polys_b.end(); ++bit)
		{
			delete *bit;
		}
		polys_b.clear();
		result.splice(result.end(),polys_a);
	}
	// for union, try to merge as much as possible
	else if(op == Union)
	{
		// first, put all in polys_a
		polys_a.splice(polys_a.end(),polys_b);
		// then for every polygon in polys_a, try to merge it with any other polygon in polys_a
		// since we know that all polygons originally in polys_a are disjunct, we don't need to
		// try to merge those
		for(std::list<Polygon *>::iterator ait = polys_a.begin(); ait != polys_a.end(); )
		{
			bool gotonext = false;

			std::list<Polygon *>::iterator bit(ait);
			for(bit++; bit != polys_a.end(); ++bit)
			{
				std::list<Polygon *> clipResult;
				Clip_result cr = (*ait)->clip(**bit,Union,clipResult);
				// delete polygons if something has been clipped and add the new ones
				if(cr == Default)
				{
					polys_a.splice(polys_a.end(),clipResult);
					delete *ait;
					delete *bit;
					polys_a.erase(bit);
					ait = polys_a.erase(ait);
					gotonext = true;
					break;
				}
				// if only clip polygon is left -> put clip polygon in polys_a
				// and delete the subject polygon
				else if(cr == ClipPolygon)
				{
					delete *ait;
					ait = polys_a.erase(ait);
					gotonext = true;
					break;
				}
				// if only subject polygon is left, simply delete the clip polygon
				// and continue
				else if(cr == SubjectPolygon)
				{
					delete *bit;
					bit = polys_b.erase(bit);
				}
			}
			if(!gotonext) ait++;
		}
		result.splice(result.end(),polys_a);
	}
	// for intersect, intersect each in A with each in B
	else if(op == Intersect)
	{
		std::list<std::list<Polygon *>::iterator> noclear_a;
		std::list<std::list<Polygon *>::iterator> noclear_b;

		for(std::list<Polygon *>::iterator ait = polys_a.begin(); ait != polys_a.end(); ait++)
		{
			for(std::list<Polygon *>::iterator bit = polys_b.begin(); bit != polys_b.end(); bit++)
			{
				std::list<Polygon *> clipResult;
				Clip_result cr = (*ait)->clip(**bit,Intersect,clipResult);
				// has been clipped. Add the new polygon to the result set
				if(cr == Default)
				{
					result.splice(result.end(),clipResult);
				}
				// an original polygon is the result. Add it to the result set and mark it
				// for non-deletion. For the subject polygon...
				else if(cr == SubjectPolygon)
				{
					result.push_back(*ait);
					// perhaps it is already inside
					if(noclear_a.back() != ait)
						noclear_a.push_back(ait);
				}
				// ... and the clip polygon
				else if(cr == ClipPolygon)
				{
					result.push_back(*bit);
					noclear_b.push_back(bit);
				}
			}
		}
		// clear the polygons in polys_a and polys_b (and delete them) except
		// for the polygons we marked
		std::list<Polygon *>::iterator it;
		std::list<std::list<Polygon *>::iterator>::iterator noit;
		for(it = polys_a.begin(), noit = noclear_a.begin(); it != polys_a.end(); ++it)
		{
			if(it == *noit)
				noit++;
			else
				delete *it;
		}
		for(it = polys_b.begin(), noit = noclear_b.begin(); it != polys_b.end(); ++it)
		{
			if(it == *noit)
				noit++;
			else
				delete *it;
		}
		polys_a.clear();
		polys_b.clear();
	}
	// for exclusion
	else if(op == Exclusion)
	{
		for(std::list<Polygon *>::iterator ait = polys_a.begin(); ait != polys_a.end(); ait++)
		{
			for(std::list<Polygon *>::iterator bit = polys_b.begin(); bit != polys_b.end(); bit++)
			{
				//std::list<Polygon *> clipResult;
				//Clip_result cr = (*ait)->clip(**bit,Exclusion,clipResult);

				// TODO
			}
		}
	}

	// TODO: to be precise, it should also return something else than Default

	return Default;
}

Polygon* PolygonClipper::createPolygon(const std::list<PolyVertex>& poly) const
{
	std::vector<Vertex> construct;
	construct.reserve(poly.size());
	for(const_polyvertex_it it = poly.begin(); it != poly.end(); ++it)
	{
		construct.push_back(it->data);
	}
	return new Polygon(construct);
}