/*
 * 2D Polygon Landscape
 *
 * Copyright (c) 2011  Tobias Zwick
 */

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <string>
#include <vector>
#include <list>
#include <set>
#include <cstdio>
#include <ctime>
#define _USE_MATH_DEFINES
#include <cmath>

#include "Vertex.h"
#include "Line.h"
#include "Polygon.h"



void testVertex()
{
	// constructor, equal, not equal
	Vertex a(10,10);
	Vertex b(a);
	Vertex c(5,10);
	assert(a == b);
	assert(a != c);
	// x() and y()
	assert(a.x() == 10);
	assert(a.y() == 10);
	a.x(5);
	assert(a == c);
	// +=, -=, +,-
	a += b;
	assert(a == Vertex(15,20));
	a -= Vertex(5,5);
	assert(a == Vertex(10,15));
	a = a+b-c;
	assert(a == Vertex(15,15));
	// *,/, *=, /=
	assert(a/3 == Vertex(5,5));
	a /= 3;
	assert(a == Vertex(5,5));
	a *= -5;
	assert(a == Vertex(-25,-25));
	assert(a*2 == Vertex(-50,-50));
	// -, +
	assert(a == +a);
	assert(-a == Vertex(25,25));
	// Skalar product
	assert(-a * Vertex(10,3) == 250+75);
}

void testLineChecks()
{
	// constructor, equal, not equal
	Line a(Vertex(0,0),Vertex(100,100));
	Line b(a);
	assert(a == b);
	b = a;
	assert(a == b);
	assert(a != Line(Vertex(0,0),Vertex(101,100)));
	// -operator
	assert(-a == Line(Vertex(100,100),Vertex(0,0)));
	// angle checks
	Line c(Vertex(345,234),Vertex(345,-100));
	assert(a.getAngle(c) == 135.0*2.0*M_PI/360.0);
	assert(c.getAngle(a) == -135.0*2.0*M_PI/360.0);
	// type
	a.setType(Line::Ray);
	assert(a != b);
	assert(a.getType() == Line::Ray);
	// vector
	Line d(Vertex(234,123),Vertex(234,1230));
	assert(d.getVector() == Vertex(0,1230-123));
	// length
	assert(a.getLength() == sqrt((double) (100*100+100*100)));
	assert(c.getLength() == sqrt((double) (0*0+334*334)));
	assert(d.getLength() == sqrt((double) (0*0+1107*1107)));
}

void testLinePointContainment()
{
	// test: containment
	Line a(Vertex(-20,-20),Vertex(20,20),Line::Segment);
	double pv;
	assert(a.contains(Vertex(0,0),pv));
	assert(pv == 0.5);
	assert(!a.contains(Vertex(0,1)));
	assert(a.contains(Vertex(-20,-20)));
	assert(a.getPoint(0.5) == Vertex(0,0));
	// test: direction
	assert(a.getDirection(Vertex(0,0)) == 0);
	assert(a.getDirection(Vertex(0,1)) == -1);
	assert(a.getDirection(Vertex(1,0)) == +1);
	// test: normals
	assert(a.getNormal() == Vertex(a.getVector().y(),-a.getVector().x()) );
}

void testLineIntersection()
{
	// test: out of range
	Line a(Vertex(-20,0),Vertex(20,0),Line::Segment);
	Line b(Vertex(40,0),Vertex(30,0),Line::Segment);
	Line c(Vertex(25,0),Vertex(29,0),Line::Ray);
	Line d(Vertex(-250,0),Vertex(-290,0),Line::Normal);

	assert(!a.intersects(b));
	assert(!b.intersects(a));
	assert(!c.intersects(a));
	assert(!a.intersects(c));

	Intersection intersect;

	// in range test, line intersections and orientation:
	// ray with line segment (but in directon of ray)
	assert(c.intersects(b,intersect));
	assert(intersect.isLineIntersection());
	assert(*intersect.line == -b);
	// line segment with ray (in direction of line)
	assert(b.intersects(c,intersect));
	assert(intersect.isLineIntersection());
	assert(*intersect.line == +b);
	// line segment with line (in direction of line)
	assert(b.intersects(d,intersect));
	assert(intersect.isLineIntersection());
	assert(*intersect.line == b);
	// line with ray
	assert(d.intersects(c,intersect));
	assert(intersect.isLineIntersection());
	assert(*intersect.line == c);
	assert(c.intersects(d,intersect));
	assert(intersect.isLineIntersection());
	assert(*intersect.line == c);

	// line intersection at end point
	b = Line(Vertex(20,0),Vertex(30,0),Line::Segment);
	assert(b.intersects(a,intersect));
	assert(!intersect.isLineIntersection());
	assert(*intersect.point == Vertex(20,0));

	// line intersection between two line segments
	a = Line(Vertex(0,0),Vertex(200,4),Line::Segment);
	b = Line(Vertex(50,1),Vertex(500,10),Line::Segment);
	assert(a.intersects(b,intersect));
	assert(intersect.isLineIntersection());
	assert(*intersect.line == Line(Vertex(50,1),Vertex(200,4),Line::Segment));
}

void testLineIntersectionAtEndpoint()
{
	// test: vertical line intersects at vertices of line
	Line a(Vertex(20,-20),Vertex(30,40),Line::Segment);
	Line b(Vertex(30,40),Vertex(40,20),Line::Segment);

	Line c(Vertex(30,-40),Vertex(30,60),Line::Segment);

	Intersection intersect;

	assert(c.intersects(a,intersect));
	assert(intersect.exact);
	assert(*intersect.point == Vertex(30,40));
	assert(c.intersects(b,intersect));
	assert(intersect.exact);
	assert(*intersect.point == Vertex(30,40));
	assert(a.intersects(b,intersect));
	assert(intersect.exact);
	assert(*intersect.point == Vertex(30,40));

	// test: line intersects at vertices of line
	a = Line(Vertex(148,-179),Vertex(141,-195),Line::Segment);
	b = Line(Vertex(141,-195),Vertex(141,-211),Line::Segment);

	c = Line(Vertex(85,-140),Vertex(180,-234),Line::Segment);

	assert(c.intersects(b,intersect));
	assert(!intersect.exact);
	assert(*intersect.point == Vertex(141,-195));
	assert(!c.intersects(a,intersect));
}

#define length(a) ( sizeof ( a ) / sizeof ( *a ) )

void clearPolygonSet(std::list<Polygon *> &set)
{
	for(std::list<Polygon *>::iterator it = set.begin(); it != set.end(); ++it)
	{
		delete *it;
	}
	set.clear();
}

Polygon createPolygon(int arr[], int length)
{
	std::list<Vertex> a;
	for(int i=0; i<length/2;i++)
	{
		a.push_back(Vertex(arr[i*2],arr[i*2+1]));
	}
	return Polygon(a);
}

bool samePolygon2D(Polygon &p, int arr[], int length)
{
	if(length%2 == 1) return false;
	Polygon q(createPolygon(arr,length));
	return p == q;
}

bool greaterOrEqualThan(const std::list<Polygon *>& set2, std::list<Polygon *> set1)
{
	// subtract all polygons in set2 from all polygons in set1
	// if outcome is empty, set2 >= set1
	for(std::list<Polygon *>::iterator it = set1.begin(); it != set1.end(); )
	{
		bool cont = true;
		for(std::list<Polygon *>::const_iterator jt = set2.begin(); jt != set2.end(); ++jt)
		{
			std::list<Polygon *> result;
			Clip_result cr = (*it)->clip(**jt,Difference,result);
			if(cr != SubjectPolygon)
			{
				set1.splice(set1.end(),result);
				it = set1.erase(it);
				cont = false;
				if(set1.empty()) return true;
				break;
			}
		}
		if(cont) ++it;
	}
	return set1.empty();
}

bool samePolygonSet(std::list<Polygon *> set1, std::list<Polygon *> set2)
{
	if(greaterOrEqualThan(set1,set2))
		if(greaterOrEqualThan(set2,set1))
			return true;
	return false;
}

Polygon getPIPExamplum()
{
	int a[] = {	94,152,		27,173,		27,97,		-14,71,		-95,113,		-184,113,
							-158,46,	-69,46,		11,38,		62,-18,		-42,9,			-149,-51,
							-54,-110,	23,-110,	127,-28,	87,59,		153,99
						};
	return createPolygon(a,length(a));
}

Polygon getPolyClipExamplumA()
{
	int a[]	= {	4,164,			28,102,			342,102,		375,211,		326,367,
							273,344,		245,150,		200,350,		100,150
						};
	return createPolygon(a,length(a));
}

Polygon getPolyClipExamplumB()
{
	int a[]	= {	261,102,		167,102,		227,9,		378,6,		411,195,		400,300,
							150,250,		90,130,			315,235,	342,102,	283,47
						};	
	return createPolygon(a,length(a));
}

Polygon getWeaklyPolyClipExamplumA()
{
	int extreme[] = {	122,9,		159,73,		115,104,
										131,52,		106,45,		115,104,
										66,104,		94,64,		70,45,
										66,104,		50,53,		31,70,
										66,104,		122,146,	35,145,
										66,104,		13,115,		16,44
									};

	return createPolygon(extreme,length(extreme));
}

Polygon getWeaklyPolyClipExamplumB()
{
	int extreme2[] = {	115,104,	119,32,		198,59,
											127,200,		7,190,		6,22,
											60,13,		66,104,		11,136,
											55,179,		66,104
									};

	return createPolygon(extreme2,length(extreme2));
}

Polygon getCaro()
{
	int a[] = {	-5,0,		-3,-2,		-2,-3,		0,-5,		5,0,		0,5	};
	return createPolygon(a,length(a));
}

Polygon getMiniCaro()
{
	int a[] = {	-2,0,		0,-2,		2,0,		0,2	};
	return createPolygon(a,length(a));
}

Polygon getQuad()
{
	int a[] = {	-5,-5,		5,-5,		5,5,		-5,5	};
	return createPolygon(a,length(a));
}

Polygon getRQuad()
{
	int a[] = {	-5,-5,		-5,5,		5,5,		5,-5	};
	return createPolygon(a,length(a));
}

Polygon getConcavum()
{
	int a[] = {	-2,0,		-5,-8,		5,-8,		2,0,		5,8,		-5,8	};
	return createPolygon(a,length(a));
}

void testTestSetup()
{
	Polygon p(getQuad());
	int a[] = {	-5,-5,		5,-5,		5,5,		-5,5	};
	assert(samePolygon2D(p,a,length(a)));
}

void testPolygonChecks()
{
	// test constructors
	Vertex arr[] = {Vertex(0,0), Vertex(10,10), Vertex(0,10) };
	std::list<Vertex> list;
	list.push_back(arr[0]);
	list.push_back(arr[1]);
	list.push_back(arr[2]);

	Polygon polyArr(arr,length(arr));
	Polygon polyRef(polyArr);
	Polygon polyLis(list);
	Polygon polyAss(arr,length(arr));
	polyAss = polyRef;
	assert(polyArr == polyArr);
	assert(polyArr == polyRef);
	assert(polyArr == polyLis);
	assert(polyArr == polyAss);

	// test size, convexity: quad
	Polygon pa(getQuad());
	assert(pa.convex());
	assert(pa.size() == 100.0);
	assert(pa.perimeter() == 40.0);

	// test size, convexity: hole quad
	Polygon pr(getRQuad());
	assert(!pr.convex());
	assert(pr.size() == -100.0);
	assert(pr.perimeter() == 40.0);

	// test size, convexity: concave polygon
	Polygon pb(getConcavum());
	assert(!pb.convex());
	assert(pb.size() == 112.0);
	assert(pb.perimeter() == 20.0+4.0*sqrt((double) (3*3+8*8)));

	// ==operator
	int a[] = {12,50,	10,10,	77,0,		0,-30};
	int b[] = {12,50,	10,10,	77,0 };
	int c[] = {12,50,	10,10,	77,0,		0,-31 };
	Polygon foo(createPolygon(a,length(a)));
	Polygon bar(createPolygon(b,length(b)));
	Polygon muh(createPolygon(c,length(c)));
	assert(foo != bar);
	assert(bar != foo);
	assert(foo != muh);
	assert(bar != muh);

	// two absolutely same simple polygons
	int a1[] = {	-5,-5,		5,-5,		5,5,		-5,5	};
	Polygon pa1(createPolygon(a1,length(a1)));
	Polygon pb1(createPolygon(a1,length(a1)));
	assert(pa1 == pb1);

	// two same polygons with different start point
	int b1[] = {		5,-5,		5,5,		-5,5,		-5,-5	};
	Polygon pc1(createPolygon(b1,length(b1)));
	assert(pc1 == pa1);

	// stress-test start_it (x,y,z are the same)
	int x[] = {	0,2,	2,4,	1,4,	0,2,	0,5,	3,6,	3,0	};
	int y[] = {	0,2,	0,5,	3,6,	3,0,	0,2,	2,4,	1,4	};
	int z[] = {	0,5,	3,6,	3,0,	0,2,	2,4,	1,4,	0,2	};
	Polygon px(createPolygon(x,length(x)));
	Polygon py(createPolygon(y,length(y)));
	Polygon pz(createPolygon(z,length(z)));
	assert(px == py);
	assert(py == px);
	assert(px == pz);
	assert(pz == px);
	assert(py == pz);
	assert(pz == py);
}

void testPointInPolygon()
{
	// tests for convex:
	Polygon pa(getCaro());
	// outside of bounding box
	assert(!pa.contains(Vertex(-6,0)));
	// inside of bounding box but outside of polygon
	assert(!pa.contains(Vertex(-3,-3)));
	// on boundary = inside
	assert(pa.contains(Vertex(-3,-2)));
	// somewhere inside
	assert(pa.contains(Vertex(2,2)));

	// tests for concave with the example from the thesis:
	Polygon po(getPIPExamplum());
	// inside (p)
	assert(po.contains(Vertex(-80,70)));
	// outside (q)
	assert(!po.contains(Vertex(-20,150)));
	// on boundary (r)
	assert(po.contains(Vertex(0,-110)));
	// outside but ray intersects vertex (a0)
	assert(!po.contains(Vertex(-149,0)));
	// inside but ray intersects vertex (b)
	assert(po.contains(Vertex(-54,-50)));
	// with line intersection (c)
	assert(!po.contains(Vertex(27,200)));
	//inside but ray intersects two vertices (a1)
	assert(po.contains(Vertex(87,100)));
}

void testPolygonLineIntersection()
{
	Polygon pa(getPIPExamplum());
	// test - concave
	std::list<Intersection> intersect;
	assert(pa.intersects(Line(Vertex(-149,0),Vertex(-149,-200)),intersect,false));
	assert(intersect.size() == 3);
	std::set<Vertex,VertexComp> results;
	results.insert(Vertex(-149,-51));
	results.insert(Vertex(-149,46));
	results.insert(Vertex(-149,113));
	for(std::list<Intersection>::iterator it = intersect.begin(); it != intersect.end(); ++it)
	{
		// unless everywhere integers are used, this value is not always correct
		// assert(intersect[i].exact);
		assert(results.find(*(it->point)) != results.end());
	}
	// test convex
	Polygon pb(getCaro());
	intersect.clear();
	// test multiple line intersections
	assert(pb.intersects(Line(Vertex(-8,3),Vertex(2,-7)),intersect,false));
	// three line and 1 point intersection (TODO: perhaps remove that point from result set?)
	assert(intersect.size() == 4);

	intersect.clear();
	// test normal
	assert(pb.intersects(Line(Vertex(-8,3),Vertex(8,-3)),intersect,false));
	assert(intersect.size() == 2);
	// test ray for convex
	intersect.clear();
	assert(pb.intersects(Line(Vertex(-8,3),Vertex(8,-3),Line::Ray),intersect,true));
	assert(intersect.size() == 1);
	assert(*(intersect.front().point) == Vertex(-4,1));
}

void testPolygonClippingStandardResults()
{
	// standard results...
	Polygon pa(getCaro());
	Polygon pb(getCaro());
	pb.move(Vertex(20,0));
	std::list<Polygon*> result;
	assert(pa.clip(pb, Difference, result) == SubjectPolygon);
	assert(pa.clip(pb, Union, result) == BothPolygons);
	assert(pa.clip(pb, Exclusion, result) == BothPolygons);
	assert(pa.clip(pb, Intersect, result) == Nothing);

	// same with inside bounding box
	Polygon pb2(getCaro());
	pb2.move(Vertex(6,6));
	assert(pa.clip(pb2, Difference, result) == SubjectPolygon);
	assert(pa.clip(pb2, Union, result) == BothPolygons);
	assert(pa.clip(pb2, Exclusion, result) == BothPolygons);
	assert(pa.clip(pb2, Intersect, result) == Nothing);
}

void testSimplePolygonClipping()
{
	Polygon pa(getCaro());
	Polygon pb(getCaro());
	pb.move(Vertex(5,0));
	std::list<Polygon*> result;

	// simple difference
	result = pa - pb;
	assert(result.size() == 1);
	int a[] = {	3,-2,		2,-2,		0,0,		3,3,		0,5,	-5,0,		-3,-2,		-2,-3,		0,-5	};
	assert(samePolygon2D(*(result.front()),a,length(a)));
	delete result.front();

	// simple union
	result = pa | pb;
	assert(result.size() == 1);
	int b[] = {	3,-2,		3,-3,		5,-5,		10,0,		5,5,		3,3,		0,5,		-5,0,		-3,-2,		-2,-3,		0,-5	};
	assert(samePolygon2D(*(result.front()),b,length(b)));
	delete result.front();

	// simple intersection
	result = pa & pb;
	assert(result.size() == 1);
	int c[] = {	3,-2,		5,0,		3,3,		0,0,		2,-2	};
	assert(samePolygon2D(*(result.front()),c,length(c)));
	delete result.front();

	// simple exclusion
	result = pa ^ pb;
	assert(result.size() == 2);
	int d[] = {	3,-2,		3,-3,		5,-5,		10,0,		5,5,		3,3,		5,0	};
	assert(samePolygon2D(*(result.front()),d,length(d)));
	delete result.front();
	result.pop_front();
	assert(samePolygon2D(*(result.front()),a,length(a)));
	delete result.front();
	result.pop_front();

}

void testPolygonClippingShortcuts()
{
	// identity
	Polygon pa(getPIPExamplum());
	std::list<Polygon*> result;
	assert(pa.clip(pa,Difference,result) == Nothing);
	assert(result.empty());
	assert(pa.clip(pa,Union,result) == SubjectPolygon);
	assert(result.empty());
	assert(pa.clip(pa,Exclusion,result) == Nothing);
	assert(result.empty());
	assert(pa.clip(pa,Intersect,result) == SubjectPolygon);
	assert(result.empty());
	
	// touch edges
	Polygon pb(getQuad());
	Polygon pc(getQuad());
	pc.move(Vertex(10,0));
	assert(pb.clip(pc,Difference,result) == SubjectPolygon);
	assert(result.empty());
	assert(pb.clip(pc,Intersect,result) == Nothing);
	assert(result.empty());
	int a[] = {	5,-5,		15,-5,		15,5,		5,5,		-5,5,		-5,-5	};
	assert(pb.clip(pc,Union,result) == Default);
	assert(samePolygon2D(*(result.front()),a,length(a)));
	delete result.front();
	result.clear();
	assert(pb.clip(pc,Exclusion,result) == Default);
	assert(samePolygon2D(*(result.front()),a,length(a)));
	delete result.front();
	result.clear();

	// touch vertices
	Polygon pd(getCaro());
	Polygon pe(getCaro());
	pe.move(Vertex(10,0));
	assert(pd.clip(pe,Difference,result) == SubjectPolygon);
	assert(result.empty());
	assert(pd.clip(pe,Union,result) == BothPolygons);
	assert(result.empty());
	assert(pd.clip(pe,Intersect,result) == Nothing);
	assert(result.empty());
	assert(pd.clip(pe,Exclusion,result) == BothPolygons);
	assert(result.empty());
}

// difference a-b
void testPolygonClippingDifferenceAB() {

	Polygon pa(getPolyClipExamplumA());
	Polygon pb(getPolyClipExamplumB());

	std::list<Polygon*> result = pa - pb;

	int a_b1[]	=	{	100,150,		4,164,			28,102,		167,102,			261,102,
									342,102,		315,235,		253,206,		245,150,		234,197,
									90,130
								};
	int a_b2[]	= {	150,250,	219,264,	200,350	};
	int a_b3[]	=	{	263,273,	350,290,	326,367,	273,344	};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(a_b3,length(a_b3))));
	assert_result.push_back(new Polygon(createPolygon(a_b1,length(a_b1))));
	assert_result.push_back(new Polygon(createPolygon(a_b2,length(a_b2))));

	assert(samePolygonSet(assert_result,result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// difference b-a
void testPolygonClippingDifferenceBA() {

	Polygon pa(getPolyClipExamplumA());
	Polygon pb(getPolyClipExamplumB());

	std::list<Polygon*> result = pb - pa;

	int b_a1[]	=	{	167,102,		227,9,		378,6,			411,195,		400,300,
									350,290,		375,211,	342,102,		283,47,			261,102
								};
	int b_a2[]	=	{	219,264,		234,197,		253,206,		263,273	};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(b_a1,length(b_a1))));
	assert_result.push_back(new Polygon(createPolygon(b_a2,length(b_a2))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// intersection
void testPolygonClippingIntersection() {

	Polygon pa(getPolyClipExamplumA());
	Polygon pb(getPolyClipExamplumB());

	std::list<Polygon*> result = pa & pb;

	int aANDb1[] =  {	100,150,		90,130,		234,197,		219,264,		150,250	};
	int aANDb2[] =  {	253,206,		315,235,	342,102,		375,211,		350,290,
										263,273
									};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(aANDb2,length(aANDb2))));
	assert_result.push_back(new Polygon(createPolygon(aANDb1,length(aANDb1))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// exclusion
void testPolygonClippingExclusion() {

	Polygon pa(getPolyClipExamplumA());
	Polygon pb(getPolyClipExamplumB());

	std::list<Polygon*> result = pa ^ pb;

	// NOTE: 1 v-hole
	int aXORb1[] = {	100,150,		4,164,		28,102,		167,102,	227,9,		378,6,
										411,195,		400,300,	350,290,	375,211,	342,102,	283,47,
										261,102,		342,102,	315,235,	253,206,	245,150,	234,197,
										90,130
									};
	int a_b2[]	= {	150,250,	219,264,	200,350	};
	int a_b3[]	=	{	263,273,	350,290,	326,367,	273,344	};
	int b_a2[]	=	{	219,264,	234,197,	253,206,	263,273	};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(a_b2,length(a_b2))));
	assert_result.push_back(new Polygon(createPolygon(a_b3,length(a_b3))));
	assert_result.push_back(new Polygon(createPolygon(b_a2,length(b_a2))));
	assert_result.push_back(new Polygon(createPolygon(aXORb1,length(aXORb1))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// union
void testPolygonClippingUnion() {

	Polygon pa(getPolyClipExamplumA());
	Polygon pb(getPolyClipExamplumB());

	std::list<Polygon*> result = pa | pb;

	int aORb1[] = {	242,9,		378,6,		405,161,		342,102,		283,47	};

	int aORb2[] = {	245,150,		234,197,		253,206,		263,273,		219,264,
									200,350,		150,250,		100,150,		4,164,			28,102,
									167,102,		225,12
								};

	int aORb3[] = {	283,47,			261,102,		342,102,		405,161,		411,195,		
									400,300,		350,290,		326,367,		273,344,		263,273,		
									253,206,		245,150,		225,12,			227,9,			242,9
								};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(aORb1,length(aORb1))));
	assert_result.push_back(new Polygon(createPolygon(aORb2,length(aORb2))));
	assert_result.push_back(new Polygon(createPolygon(aORb3,length(aORb3))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// difference a-b
void testWeaklyPolygonClippingDifferenceAB()
{
	Polygon pa(getWeaklyPolyClipExamplumA());
	Polygon pb(getWeaklyPolyClipExamplumB());

	std::list<Polygon*> result = pa - pb;
	
	int aDiffb1[] = {	60,145,		35,145,		66,104	};
	// NOTE: 1 v-hole
	int aDiffb2[] = {	61,29,		122,9,		139,39,		119,32,		118,48,		106,45,
										115,104,	66,104,		94,64,		70,45,		66,104	
									};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(aDiffb1,length(aDiffb1))));
	assert_result.push_back(new Polygon(createPolygon(aDiffb2,length(aDiffb2))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// difference b-a
void testWeaklyPolygonClippingDifferenceBA()
{
	Polygon pa(getWeaklyPolyClipExamplumA());
	Polygon pb(getWeaklyPolyClipExamplumB());

	std::list<Polygon*> result = pb - pa;

	int bDiffa1[] = {	60,145,		122,146,		66,104,		115,104,		159,73,
										139,39,		198,59,			127,200,		7,190,		6,22,
										60,13,		61,29,			16,44,		13,115,			66,104,
										11,136,		55,179
									};

	int bDiffa2[] = {	66,104,		31,70,		50,53		};
	int bDiffa3[] = {	115,104,	118,48,		131,52	};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(bDiffa1,length(bDiffa1))));
	assert_result.push_back(new Polygon(createPolygon(bDiffa2,length(bDiffa2))));
	assert_result.push_back(new Polygon(createPolygon(bDiffa3,length(bDiffa3))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// union
void testWeaklyPolygonClippingUnion()
{
	Polygon pa(getWeaklyPolyClipExamplumA());
	Polygon pb(getWeaklyPolyClipExamplumB());

	std::list<Polygon*> result = pa | pb;

	// NOTE: 1 v-hole
	int union1[] = {	7,132,		11,136,		55,179,		72,195,		7,190	};

	int union2[] = {	11,21,		60,13,		61,29,		122,9,		139,39,		
										198,59,		194,67,		118,48,		106,45
									};

	int union3[] = {	106,45,		115,104,	118,48,		194,67,		127,200,		
										72,195,		55,179,		60,145,		35,145,		66,104,		
										94,64,		70,45,		66,104,		11,136,		7,132,		
										6,22,			11,21
								};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(union1,length(union1))));
	assert_result.push_back(new Polygon(createPolygon(union2,length(union2))));
	assert_result.push_back(new Polygon(createPolygon(union3,length(union3))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// intersection
void testWeaklyPolygonClippingIntersection()
{
	Polygon pa(getWeaklyPolyClipExamplumA());
	Polygon pb(getWeaklyPolyClipExamplumB());
	
	std::list<Polygon*> result = pa & pb;

	int intersection1[] = {	60,145,		66,104,		122,146	};
	// NOTE: 1 v-hole
	int oddPolyWHole[] =	{	61,29,		66,104,		50,53,		31,70,		66,104,
													13,115,		16,44
												};

	int intersection2[] = {	115,104,		131,52,		118,48,		119,32,		139,39,
													159,73
												};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(intersection1,length(intersection1))));
	assert_result.push_back(new Polygon(createPolygon(intersection2,length(intersection2))));
	assert_result.push_back(new Polygon(createPolygon(oddPolyWHole,length(oddPolyWHole))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// exclusion
void testWeaklyPolygonClippingExclusion()
{
	Polygon pa(getWeaklyPolyClipExamplumA());
	Polygon pb(getWeaklyPolyClipExamplumB());
	
	std::list<Polygon*> result = pa ^ pb;

	// NOTE: this weakly simple polygons has 6 v-holes and 3 v-polys
	int wow[] = {	60,145,		122,146,		66,104,		94,64,		70,45,		66,104,
								61,29,		122,9,			139,39,		119,32,
								118,48,		131,52,			115,104,	118,48,		// a little loop here
								106,45,		115,104,		159,73,		139,39,		198,59,		127,200,
								7,190,		6,22,				60,13,		61,29,		16,44,		13,115,
								66,104,		31,70,			50,53,		66,104,	// another loop here
								11,136,		55,179,			60,145,
								35,145,		66,104												// and third loop here
							};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(wow,length(wow))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}


void testRecursivePolygonIntersectionFinding()
{
	int one[] = {	6,13,		16,4,		21,29,		10,32,		6,15	};
	int two[] = {	6,15,		9,10,		19,1,			27,17,		10,10,		21,29	};

	Polygon pa(createPolygon(one,length(one)));
	Polygon pb(createPolygon(two,length(two)));

	std::list<Polygon *> result = pa - pb;

	int result1[] = {	6,15,		6,13,		9,11	};
	int result2[] = {	6,15,		21,29,	10,32	};
	int result3[] = {	10,10,	18,13,	21,29	};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(result2,length(result2))));
	assert_result.push_back(new Polygon(createPolygon(result1,length(result1))));
	assert_result.push_back(new Polygon(createPolygon(result3,length(result3))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

void testPolygonSplittingSimple()
{
	Polygon polygon(getPolyClipExamplumB());

	// with vertical line
	std::list<Polygon *> result = polygon / Line(Vertex(283,50),Vertex(283,0),Line::Segment);
	assert(result.size() == 3);

	int divide0_1[] = {	283,8,		378,6,			411,195,		400,300,		283,277,
											283,220,	315,235,		342,102,		283,47
										};
	int divide0_2[] = {	283,8,		283,47,			261,102,		167,102,		227,9	};
	int divide0_3[] = {	283,220,	283,277,		150,250,		90,130	};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(divide0_1,length(divide0_1))));
	assert_result.push_back(new Polygon(createPolygon(divide0_2,length(divide0_2))));
	assert_result.push_back(new Polygon(createPolygon(divide0_3,length(divide0_3))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

void testPolygonSplittingOther()
{
	Polygon polygon(getPolyClipExamplumB());
	std::list<Polygon *> result;

	// with line outside bounding box
	assert(polygon.split(Line(Vertex(250,-250),Vertex(500,0)),result) == SubjectPolygon);
	// with line inside bounding box but outside polygon
	assert(polygon.split(Line(Vertex(226,9),Vertex(90,129)),result) == SubjectPolygon);
	// with line intersecting with bounding box only in a line
	assert(polygon.split(Line(Vertex(90,0),Vertex(90,131)),result) == SubjectPolygon);
	// with line intersecting with bounding box only in a vertex
	assert(polygon.split(Line(Vertex(405,0),Vertex(411,6)),result) == SubjectPolygon);
}

void testPolygonSplittingNormal()
{
	Polygon polygon(getPolyClipExamplumB());

	// with normal line
	std::list<Polygon *> result = polygon / Line(Vertex(150,240),Vertex(342,95));
	assert(result.size() == 3);

	int divide1_1[] = {	146,243,	90,130,		217,189,	150,240	};


	int divide1_2[] = {	338,98,		283,47,		261,102,		167,102,
											227,9,		378,6,		388,61,			342,95
										};

	int divide1_3[] = {	146,243,	150,240,	217,189,		315,235,		342,102,
											338,98,		342,95,		388,61,			411,195,		400,300,
											150,250
										};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(divide1_1,length(divide1_1))));
	assert_result.push_back(new Polygon(createPolygon(divide1_2,length(divide1_2))));
	assert_result.push_back(new Polygon(createPolygon(divide1_3,length(divide1_3))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// with normal line
// float rounding imprecision could lead to that one intersection is
// not found
void testPolygonSplittingImprecision()
{
	Polygon polygon(getPolyClipExamplumB());
	std::list<Polygon *> result = polygon / Line(Vertex(150,250),Vertex(342,102));
	assert(result.size() == 3);

	int divide2_1[] = {	150,250,		90,130,		224,193	};

	int divide2_2[] = {	342,102,		283,47,		261,102,		167,102,		
											227,9,			378,6,		389,66
										};

	int divide2_3[] = {	150,250,		224,193,	315,235,		342,102,		
											389,66,		411,195,		400,300
										};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(divide2_1,length(divide2_1))));
	assert_result.push_back(new Polygon(createPolygon(divide2_2,length(divide2_2))));
	assert_result.push_back(new Polygon(createPolygon(divide2_3,length(divide2_3))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// and another one (along reverse shared edge)
void testPolygonSplittingRSEdge()
{
	Polygon polygon(getPolyClipExamplumB());

	std::list<Polygon *> result = polygon / Line(Vertex(227,9),Vertex(167,102));
	assert(result.size() == 2);

	int divide3_2[] = {	116,181,		90,130,		135,151	};

	int divide3_1[] = {	116,181,		135,151,		315,235,		342,102,		283,47,
											261,102,		167,102,		227,9,			378,6,
											411,195,		400,300,		150,250
										};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(divide3_2,length(divide3_2))));
	assert_result.push_back(new Polygon(createPolygon(divide3_1,length(divide3_1))));

	assert(samePolygonSet(result,assert_result));

	clearPolygonSet(result);
	clearPolygonSet(assert_result);
}

// all holes results
void testHoles()
{
	Polygon pa(getCaro());
	Polygon pb(getMiniCaro());

	std::list<Polygon *> result;

	assert(pb.clip(pa,Difference,result) == Nothing);
	assert(pa.clip(pb,Union,result) == SubjectPolygon);
	assert(pb.clip(pa,Union,result) == ClipPolygon);
	assert(pa.clip(pb,Intersect,result) == ClipPolygon);
	assert(pb.clip(pa,Intersect,result) == SubjectPolygon);

	int hole_diff1[] = {	0,-2,		-2,0,		0,2,		0,5,		-5,0,		0,-5	};
	int hole_diff2[] = {	0,-2,		0,-5,		5,0,		0,5,		0,2,		2,0		};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(hole_diff1,length(hole_diff1))));
	assert_result.push_back(new Polygon(createPolygon(hole_diff2,length(hole_diff2))));

	assert(pa.clip(pb,Difference,result) == Default);
	assert(samePolygonSet(result,assert_result));
	clearPolygonSet(result);

	assert(pa.clip(pb,Exclusion,result) == Default);
	assert(samePolygonSet(result,assert_result));
	clearPolygonSet(result);

	clearPolygonSet(assert_result);
}

void testSimplePolygonTurnsIntoComplexPolygon()
{
	// taken from actual crash data
	int c[] =	{	187,-323,		230,-239,		227,-243	};
	int d[] = {	401,-369,		65,-230,		65,-370,		401,-370 };

	Polygon pc(createPolygon(c,length(c)));
	Polygon pd(createPolygon(d,length(d)));

	pc - pd;

	// TODO: finish the test for pc and pd.

	// taken from actual crash data (-340,+190)
	int a[] = {	16,-7,		9,-9,			5,-8,		8,-9,		12,-10,		19,-8	};
	int b[] = {	23,-14,		16,-10,		9,-8,		2,-10	};

	Polygon pa(createPolygon(a,length(a)));
	Polygon pb(createPolygon(b,length(b)));

	// what it should look like in the end (difference)
	int result1[] = {	5,-8,		7,-9,		8,-9	};
	int result2[] = {	11,-8,	14,-9,	19,-8,		16,-7	};

	std::list<Polygon*> assert_result;
	assert_result.push_back(new Polygon(createPolygon(result1,length(result1))));
	assert_result.push_back(new Polygon(createPolygon(result2,length(result2))));

	std::list<Polygon *> result = pa - pb;

	assert(samePolygonSet(result,assert_result));

	// TODO: this test should test all edge cases of that:
	// lines, negative polygons, connected polygons through "lines" (like a chain),
	// flower-like polygons
}

int main(int argc, char *argv[])
{

	time_t start = time(NULL);

	// Vertex tests
	testVertex();

	// Line tests
	testLineChecks();
	testLinePointContainment();
	testLineIntersection();
	testLineIntersectionAtEndpoint();

	// Polygon tests
	testPolygonChecks();
	testPointInPolygon();
	testPolygonLineIntersection();

	// Polygon clipping (basics)
	testTestSetup();
	testPolygonClippingStandardResults();
	testSimplePolygonClipping();
	testPolygonClippingShortcuts();

	// normal Polygon clipping
	testPolygonClippingDifferenceAB();
	testPolygonClippingDifferenceBA();
	testPolygonClippingExclusion();
	testPolygonClippingIntersection();
	testPolygonClippingUnion();

	// weakly polygon clipping
	// TODO: weakly simple polygon clipping is currently broken because of the 
	// incomplete self-intersection handling. But weakly simple polygons were
	// not part of the bachelor thesis anyway
	/*
	testWeaklyPolygonClippingDifferenceAB();
	testWeaklyPolygonClippingDifferenceBA();
	testWeaklyPolygonClippingExclusion();
	testWeaklyPolygonClippingIntersection();
	testWeaklyPolygonClippingUnion();
*/
	// splitting
	testPolygonSplittingSimple();
	testPolygonSplittingNormal();
	testPolygonSplittingImprecision();
	testPolygonSplittingRSEdge();

	// holes and other special cases
	testHoles();
	testRecursivePolygonIntersectionFinding();
	//testSimplePolygonTurnsIntoComplexPolygon(); //TODO

	time_t delta = time(NULL) - start;
	printf("Successfully passed %d tests in %d second(s).",23,(int)delta);
	std::cin.get();

  return 0;
}